#include"Image.h"

#include<stdio.h>
#include<algorithm>
#include<string>
#include<vector>
#include<math.h>
#include<assert.h>
#include<random>
#include<queue>

using namespace std;
using lf = double;

using namespace IMAGE;

struct vec2 {
	lf x, y;
	vec2() {
		x = y = 0;
	}
	vec2(lf x, lf y) :x(x), y(y) {}

	vec2 operator +(const vec2& z) {
		return vec2(x + z.x, y + z.y);
	}
	vec2 operator -(const vec2& z) {
		return vec2(x - z.x, y - z.y);
	}
	vec2 operator *(const lf& z) {
		return vec2(x * z, y * z);
	}
	lf operator *(const vec2& z) {
		return x * z.x + y * z.y;
	}
	void normalize() {
		lf dd = sqrt(x * x + y * y);
		x /= dd;
		y /= dd;
	}
};

Image img;
Image number[10];
Image imagebeacon;
int width = 960;
int height = 640;
int beacon_color_min = 100;
int beacon_color_max = 200;

lf beacon_rate = 0.5;
lf mountain_padding_x = 20;
lf mountain_padding_y = 20;
lf mountain_width = 4;
lf mountain_under_rr = 61.0 / 255.0 * 0.7;
lf mountain_under_gg = 237.0 / 255.0 * 0.7;
lf mountain_under_bb = 32.0 / 255.0 * 0.7;
lf village_width = 4;
lf village_rr = 0.63;
lf village_gg = 0.14;
lf village_bb = 0.29;


mt19937 rnd;
int nextint(int s, int t) {
	return (int)(rnd() % (t - s + 1)) + s;
}

void inil()
{
	int i, j, k;
	img.setsize(width, height);
	for (i = 0; i < 10; i++) {
		char xxx[99];
		sprintf(xxx, "numbers/%d.png", i);
		number[i].Load(xxx);
		//number[i].set_transparency_color(255, 255, 255, 0);
		for (int j = 0; j < number[i].h; j++) {
			for (int k = 0; k < number[i].w; k++) {
				int id = (j * number[i].w + k);
				int z = number[i].data[id * 4 + 0];
				number[i].data[id * 4 + 0] = 0;
				number[i].data[id * 4 + 1] = 0;
				number[i].data[id * 4 + 2] = 0;
				number[i].data[id * 4 + 3] = 255 - z;
			}
		}
	}
	imagebeacon.Load("beacon.png");
}
Image build_beacon() {
	int i, j, k;
	Image beacon = imagebeacon;
	int w = beacon.w;
	int h = beacon.h;
	vector<vector<bool>> vb(h, vector<bool>(w, false));
	vector<vector<lf>> vc(h, vector<lf>(w, 0.0));
	vector<int> que(w * h, 0);
	int q, r;
	int dx[4] = { 0,1,0,-1 };
	int dy[4] = { 1,0,-1,0 };
	for (i = 0; i < h; i++) {
		for (j = 0; j < w; j++) {
			int id = i * w + j;
			if (vb[i][j])continue;
			if (beacon.data[id * 4 + 3] > 0) {
				int state = 0;
				q = r = 0;
				int ssi, ssj, sei, sej;
				int esi, esj, eei, eej;
				ssi = esi = h;
				ssj = esj = h;
				sei = eei = -1;
				sej = eej = -1;
				if (beacon.data[id * 4 + 0] == 0 && beacon.data[id * 4 + 1] == 255 && beacon.data[id * 4 + 2] == 255)
				{
					state = 1;
					que[r++] = id;
					vb[i][j] = true;
					ssi = min(ssi, i);
					ssj = min(ssj, j);
					sei = max(sei, i);
					sej = max(sej, j);
				}
				if (beacon.data[id * 4 + 0] == 255 && beacon.data[id * 4 + 1] == 0 && beacon.data[id * 4 + 2] == 255)
				{
					state = 1;
					que[r++] = id;
					vb[i][j] = true;
					esi = min(esi, i);
					esj = min(esj, j);
					eei = max(eei, i);
					eej = max(eej, j);
				}
				if (beacon.data[id * 4 + 0] == 255 && beacon.data[id * 4 + 1] == 255 && beacon.data[id * 4 + 2] == 255)
				{
					state = 2;
					que[r++] = id;
					vb[i][j] = true;
				}
				while (q < r) {
					int qi = que[q] / w;
					int qj = que[q] % w;
					q++;
					for (k = 0; k < 4; k++) {
						int ti = qi + dx[k];
						int tj = qj + dy[k];
						if (ti >= 0 && ti < h && tj >= 0 && tj < w) {
							if (vb[ti][tj])continue;
							if (
								//beacon.data[(ti * w + tj) * 4 + 0] == beacon.data[(qi * w + qj) * 4 + 0] &&
								//beacon.data[(ti * w + tj) * 4 + 1] == beacon.data[(qi * w + qj) * 4 + 1] &&
								beacon.data[(ti * w + tj) * 4 + 2] == beacon.data[(qi * w + qj) * 4 + 2] &&
								beacon.data[(ti * w + tj) * 4 + 3] == beacon.data[(qi * w + qj) * 4 + 3])
							{
								vb[ti][tj] = true;
								que[r++] = ti * w + tj;
								if (beacon.data[(ti * w + tj) * 4 + 0] == 0) {
									ssi = min(ssi, ti);
									ssj = min(ssj, tj);
									sei = max(sei, ti);
									sej = max(sej, tj);
								}
								else if (beacon.data[(ti * w + tj) * 4 + 1] == 0) {
									esi = min(esi, ti);
									esj = min(esj, tj);
									eei = max(eei, ti);
									eej = max(eej, tj);
								}
							}
						}
					}
				}
				if (state == 2) {
					int color = nextint(beacon_color_min, beacon_color_max);
					for (k = 0; k < r; k++) {
						int qi = que[k] / w;
						int qj = que[k] % w;
						beacon.data[(qi * w + qj) * 4 + 0] = color;
						beacon.data[(qi * w + qj) * 4 + 1] = color;
						beacon.data[(qi * w + qj) * 4 + 2] = color;
						beacon.data[(qi * w + qj) * 4 + 3] = 255;
					}
				}
				else if (state == 1) {
					for (k = 0; k < r; k++) {
						int qi = que[k] / w;
						int qj = que[k] % w;
						int color = 0;
						if (esi <= qi && qi <= eei && esj <= qj && qj <= eej) color = 5;
						else {
							lf scale_x = 0.0;
							lf scale_y = 0.0;
							if (qi < esi) scale_x = 1.0 - (lf)(qi - ssi) / (esi - ssi);
							else if (qi > eei) scale_x = (lf)(qi - eei) / (sei - eei);
							if (qj < esj) scale_y = 1.0 - (lf)(qj - ssj) / (esj - ssj);
							else if (qj > eej) scale_y = (lf)(qj - eej) / (sej - eej);
							color = 5 + (beacon_color_min - 5) * max(scale_x, scale_y);
						}
						beacon.data[(qi * w + qj) * 4 + 0] = color;
						beacon.data[(qi * w + qj) * 4 + 1] = color;
						beacon.data[(qi * w + qj) * 4 + 2] = color;
						beacon.data[(qi * w + qj) * 4 + 3] = 255;
					}
				}
			}
		}
	}
	beacon.set_transparency_color(100, 0, 0, 0);
	return beacon;
}
int n = 5;
int a[99] = { 1, 4, 2, 5, 3 };
int range[5][2] = {
	{0, 1},
	{1, 3},
	{3, 5},
};
lf range_color[5][4] = {
	{200.0 / 255.0, 58.0 / 255.0, 142.0 / 255.0, 0.3},
	{49.0 / 255.0, 194.0 / 255.0, 238.0 / 255.0, 0.3},
	{229.0 / 255.0, 159.0 / 255.0, 1.0 / 255.0, 0.3},
};
int rl = 3;

void draw_mountain_line() {
	// padding / line / beacon / line ... / beacon / line / padding
	int i, j, k;
	lf line_width = (lf)(width - 2 * mountain_padding_x) / (n + 1 + n * beacon_rate);
	lf beacon_width = line_width * beacon_rate;
	lf beacon_height = beacon_width * imagebeacon.h / imagebeacon.w;
	lf line_height_unit = (lf)(height - 2 * mountain_padding_y - beacon_height) / n;

	lf radius = mountain_width;
	a[n] = 0;

	int prev = 0;
	lf cx, cy;
	lf px, py;
	{
		cx = -1;
		cy = height - mountain_padding_y - line_height_unit * 0;
		px = mountain_padding_x;
		py = height - mountain_padding_y - line_height_unit * 0;
		img.draw_line(cx, cy, px, py, radius, 0, 0, 0);
	}
	for (i = 0; i <= n; i++) {
		cx = mountain_padding_x + i * (line_width + beacon_width);
		cy = height - mountain_padding_y - line_height_unit * prev;
		px = mountain_padding_x + i * (line_width + beacon_width) + line_width;
		py = height - mountain_padding_y - line_height_unit * a[i];
		img.draw_line(cx, cy, px, py, radius, 0, 0, 0);
		if (i < n) {
			img.draw_line(px, py, px + beacon_width, py, radius, 0, 0, 0);
		}
		prev = a[i];
	}
	{
		cx = width - mountain_padding_x;
		cy = height - mountain_padding_y - line_height_unit * 0;
		px = width + 1;
		py = height - mountain_padding_y - line_height_unit * 0;
		img.draw_line(cx, cy, px, py, radius, 0, 0, 0);
	}
}

void draw_village_line() {
	// padding / line / beacon / line ... / beacon / line / padding
	int i, j, k;
	lf line_width = (lf)(width - 2 * mountain_padding_x) / (n + 1 + n * beacon_rate);
	lf beacon_width = line_width * beacon_rate;
	lf beacon_height = beacon_width * imagebeacon.h / imagebeacon.w;
	lf line_height_unit = (lf)(height - 2 * mountain_padding_y - beacon_height) / n;

	lf radius = village_width;
	a[n] = 0;

	int prev = 0;
	lf cx, cy;
	lf px, py;
	for (i = 0; i < n; i++) {
		px = mountain_padding_x + i * (line_width + beacon_width) + line_width;
		py = height - mountain_padding_y - line_height_unit * a[i];
		if (i < n) {
			img.draw_line(px, py, px + beacon_width, py, radius, village_rr, village_gg, village_bb);
		}
	}
}

void draw_village_number() {
	// padding / line / beacon / line ... / beacon / line / padding
	int i, j, k;
	lf line_width = (lf)(width - 2 * mountain_padding_x) / (n + 1 + n * beacon_rate);
	lf beacon_width = line_width * beacon_rate;
	lf beacon_height = beacon_width * imagebeacon.h / imagebeacon.w;
	lf line_height_unit = (lf)(height - 2 * mountain_padding_y - beacon_height) / n;

	a[n] = 0;

	int prev = 0;
	lf cx, cy;
	lf px, py;
	lf number_size = min(beacon_width, line_height_unit) * 0.9;
	lf number_offset = mountain_width;
	for (i = 0; i < n; i++) {
		px = mountain_padding_x + i * (line_width + beacon_width) + line_width + beacon_width * 0.5;
		py = height - mountain_padding_y - line_height_unit * a[i];
		img.draw_image(px - number_size * 0.5, py + number_offset, px + number_size * 0.5, py + number_offset + number_size, number[i + 1]);
	}
}
void draw_beacon()
{
	// padding / line / beacon / line ... / beacon / line / padding
	int i, j, k;
	lf line_width = (lf)(width - 2 * mountain_padding_x) / (n + 1 + n * beacon_rate);
	lf beacon_width = line_width * beacon_rate;
	lf beacon_height = beacon_width * imagebeacon.h / imagebeacon.w;
	lf line_height_unit = (lf)(height - 2 * mountain_padding_y - beacon_height) / n;

	lf cx, cy;
	lf px, py;
	for (k = 0; k < rl; k++) {
		int bi = range[k][0];
		for (i = range[k][0] + 1; i < range[k][1]; i++) {
			if (a[bi] < a[i])bi = i;
		}
		{
			cx = mountain_padding_x + bi * (line_width + beacon_width) + line_width;
			cy = height - mountain_padding_y - line_height_unit * a[bi] - beacon_height;
			px = mountain_padding_x + bi * (line_width + beacon_width) + line_width + beacon_width;
			py = height - mountain_padding_y - line_height_unit * a[bi];
			img.draw_image(cx, cy, px, py, build_beacon());
		}
	}
}
void draw_bad_beacon()
{
	// padding / line / beacon / line ... / beacon / line / padding
	int i, j, k;
	lf line_width = (lf)(width - 2 * mountain_padding_x) / (n + 1 + n * beacon_rate);
	lf beacon_width = line_width * beacon_rate;
	lf beacon_height = beacon_width * imagebeacon.h / imagebeacon.w;
	lf line_height_unit = (lf)(height - 2 * mountain_padding_y - beacon_height) / n;

	lf cx, cy;
	lf px, py;
	k = 1;
	{
		int bi = range[k][0];
		for (i = range[k][0] + 1; i < range[k][1]; i++) {
			if (a[bi] < a[i])bi = i;
		}
		{
			cx = mountain_padding_x + bi * (line_width + beacon_width) + line_width - beacon_width * 0.25;
			cy = height - mountain_padding_y - line_height_unit * a[bi] - beacon_height;
			px = mountain_padding_x + bi * (line_width + beacon_width) + line_width + beacon_width + beacon_width * 0.25;
			py = height - mountain_padding_y - line_height_unit * a[bi];
			img.draw_line(cx, cy, px, py, mountain_width, 1.0, 0.2, 0.2);
			img.draw_line(px, cy, cx, py, mountain_width, 1.0, 0.2, 0.2);
		}
	}
}
void draw_beacon_range(bool alpha_one)
{
	// padding / line / beacon / line ... / beacon / line / padding
	int i, j, k;
	lf line_width = (lf)(width - 2 * mountain_padding_x) / (n + 1 + n * beacon_rate);
	lf beacon_width = line_width * beacon_rate;
	lf beacon_height = beacon_width * imagebeacon.h / imagebeacon.w;
	lf line_height_unit = (lf)(height - 2 * mountain_padding_y - beacon_height) / n;

	lf range_line_width = line_width * 0.05;

	lf cx, cy;
	lf px, py;
	for (k = 0; k < rl; k++) {
		int bi = range[k][0];
		for (i = range[k][0] + 1; i < range[k][1]; i++) {
			if (a[bi] < a[i])bi = i;
		}
		cx = mountain_padding_x + line_width + range[k][0] * (line_width + beacon_width) - line_width * 0.5 + range_line_width;
		px = mountain_padding_x + line_width + range[k][1] * (line_width + beacon_width) - line_width * 0.5 - range_line_width;
		cy = height - mountain_padding_y - line_height_unit * a[bi] - beacon_height - range_line_width;
		// draw outside line
		img.draw_line(cx, cy, cx, height + 1, range_line_width, range_color[k][0], range_color[k][1], range_color[k][2], 1.0);
		img.draw_line(px, cy, px, height + 1, range_line_width, range_color[k][0], range_color[k][1], range_color[k][2], 1.0);
		img.draw_line(cx, cy, px, cy, range_line_width, range_color[k][0], range_color[k][1], range_color[k][2], 1.0);
		for (i = cy; i < height; i++) {
			for (j = cx - 1; j <= px + 1; j++) {
				if (j < cx + range_line_width || j > px - range_line_width)continue;
				img.set_pixel(i, j, range_color[k][0], range_color[k][1], range_color[k][2], alpha_one ? 0.5 : range_color[k][3]);
			}
		}
	}
}

void draw_mountain_under() {
	// padding / line / beacon / line ... / beacon / line / padding
	int i, j, k;
	lf line_width = (lf)(width - 2 * mountain_padding_x) / (n + 1 + n * beacon_rate);
	lf beacon_width = line_width * beacon_rate;
	lf beacon_height = beacon_width * imagebeacon.h / imagebeacon.w;
	lf line_height_unit = (lf)(height - 2 * mountain_padding_y - beacon_height) / n;

	a[n] = 0;
	int tn = 1;
	int tm = 0;
	int btm = tm * beacon_rate;
	lf ymax = height + 30;
	lf grass_range = 15;
	img.mSamplingMode = SamplingModeType::SAMPLING_3x3;
	for (k = 0; k < tn; k++) {
		vector<lf> pxy;

		lf cx, cy;
		lf px, py;
		lf scale = (lf)(tn - k) / tn;
		{
			cx = -1;
			cy = ymax - scale * (ymax - (height - mountain_padding_y - line_height_unit * 0));
			px = mountain_padding_x;
			py = ymax - scale * (ymax - (height - mountain_padding_y - line_height_unit * 0));
			pxy.push_back(cx);
			pxy.push_back(cy);
			for (j = 0; j < tm; j++) {
				lf zx = (px - cx) * (j + 0.5) / tm + cx;
				lf zy = (py - cy) * (j + 0.5) / tm + cy;
				zy += nextint(-grass_range, 0);
				pxy.push_back(zx);
				pxy.push_back(zy);
			}
			pxy.push_back(px);
			pxy.push_back(py);
			cx = px;
			cy = py;
		}
		for (i = 0; i <= n; i++) {
			px = mountain_padding_x + i * (line_width + beacon_width) + line_width;
			py = ymax - scale * (ymax - (height - mountain_padding_y - line_height_unit * a[i]));

			for (j = 0; j < tm; j++) {
				lf zx = (px - cx) * (j + 0.5) / tm + cx;
				lf zy = (py - cy) * (j + 0.5) / tm + cy;
				zy += nextint(-grass_range, 0);
				pxy.push_back(zx);
				pxy.push_back(zy);
			}
			pxy.push_back(px);
			pxy.push_back(py);
			cx = px;
			cy = py;
			if (i < n) {
				px += beacon_width;
				for (j = 0; j < btm; j++) {
					lf zx = (px - cx) * (j + 0.5) / btm + cx;
					lf zy = (py - cy) * (j + 0.5) / btm + cy;
					zy += nextint(-grass_range, 0);
					pxy.push_back(zx);
					pxy.push_back(zy);
				}
				pxy.push_back(px);
				pxy.push_back(py);
				cx = px;
				cy = py;
			}
		}
		{
			for (j = 0; j < tm; j++) {
				lf zx = (px - cx) * (j + 0.5) / tm + cx;
				lf zy = (py - cy) * (j + 0.5) / tm + cy;
				zy += nextint(-grass_range, 0);
				pxy.push_back(zx);
				pxy.push_back(zy);
			}
			px = width + 1;
			py = ymax - scale * (ymax - (height - mountain_padding_y - line_height_unit * 0));
			pxy.push_back(px);
			pxy.push_back(py);
		}
		{
			pxy.push_back(width + 1);
			pxy.push_back(ymax);
			pxy.push_back(-1);
			pxy.push_back(ymax);
		}
		img.draw_polygon(pxy.size() / 2, pxy,
			tn == 1 ? mountain_under_rr : nextint(0.0, 100.0) / 255.0,
			tn == 1 ? mountain_under_gg : nextint(150, 255) / 255.0,
			tn == 1 ? mountain_under_bb : nextint(0.0, 100.0) / 255.0,
			(k == 0 ? 1.0 : nextint(50, 100) / 100.0)
		);
	}
	img.mSamplingMode = SamplingModeType::SAMPLING_CENTER;
}
int main() {
	int i, j, k;
	inil();
	build_beacon().Save("beacon_modify.png");

	draw_beacon_range(true);
	draw_mountain_under();
	draw_beacon();
	draw_mountain_line();
	draw_village_line();
	draw_beacon_range(false);
	draw_village_number();
	img.Save("beacon_output.png");

	img.clean();
	range[0][1] = 2;
	range[1][0] = 2;
	draw_beacon_range(true);
	draw_mountain_under();
	draw_beacon();
	draw_mountain_line();
	draw_village_line();
	draw_beacon_range(false);
	draw_village_number();
	draw_bad_beacon();

	img.Save("beacon_bad_output.png");
}